(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Interpr.sml,v
   Comment: Interpreter for statecharts represented as in Absyn.sml
   Release: scope-0-15_final 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   PROBLEMS

   1. (minor) boolean expressions are eager while they are lazy in C.
      Could be done by treating them specially in evaluator (separately from
      all other eager binary primitives) and translating andx and orx to lazy
      evaluation style.
      
   2. (minor) I am not sure if shifts implemented are the right ones...
   
   3. (not so important) List of instances could be passed in CPS style.
      It is mostly passed so now. Explensive appends happen only where we now
      that the list will actually be relatively short (so the cost is
      bounded). Can easily be improved to make everyting in accumulative
      style.

   4. (????) Variable environment venv is not modified (and not passed) by
      evaluator. In practice variables may be modified in arithmetic
      expressions. Perhaps it would be a better modelling to carry venv 
      around (and pretending it is modified) 

   5. (efficiency) the whole thing could possibly be more efficient if
      orstates completely disappeared from abstract syntax. Better an IR
      without orstates could be developed, and this should be intererpreted.

   6. (FIXME) Constants are not taken into account even if they exist in
      statechart model (probably would cause a runtime error).

   Interpr.sml,v 1.26 2003/07/24 13:19:43 wasowski Exp
*)

structure Interpr = struct

(* interface to external modules *)

open Absyn SymbolEnv List
infixr 0 $;

structure S = Symbol
structure F = Fifo

type aor = unit
type atr = unit

(* --- helper brought from Absyn *)

structure StateSet = 
      SetFn (BinarySetFn( type ord_key = aor state val compare= compare ))

structure ORStateSet =
      SetFn (BinarySetFn( type ord_key = aor orstate val compare= compareor ))

structure ANDStateSet =
      SetFn (BinarySetFn( type ord_key = aor andstate val compare= compareand ))

type orstateset  = ORStateSet.set
type andstateset = ANDStateSet.set
type stateset    = StateSet.set

local
	fun dor (s :aor orstate, res :stateset) :stateset =
		let val kids = childrenor s
		    val res1 = foldl (fn(k,r)=>StateSet.add(r,InclAND k)) res kids
		in foldl dand res1 kids end

	and dand (s :aor andstate, res :stateset) :stateset = 
		let val kids= childrenand s
		    val res1= foldl (fn(k,r)=>StateSet.add(r,InclOR k)) res kids
		in foldl dor res1 kids end
  in
	fun descendor  (s :aor orstate ) :stateset = dor  (s,StateSet.empty)
	fun descendand (s :aor andstate) :stateset = dand (s,StateSet.empty)
  end

fun descendandst (s :aor andstate) :stateset =
	lift_st descendand (fn (s,ds) => StateSet.add(ds,InclAND s)) s
fun descendorst (s :aor orstate) :stateset =
	lift_st descendor  (fn (s,ds) => StateSet.add(ds,InclOR s)) s
fun descend (s :aor state) :stateset = lift_orand descendor descendand s
  
fun descendst (s :aor state) :stateset = lift_st descend StateSet.add' s

fun descendh (h :aor hierarchy) :stateid -> stateset = lift_h descend h

fun onlyand (s :stateset) :andstateset = 
	ANDStateSet.addList 
		(ANDStateSet.empty, List.mapPartial getand $ 
			StateSet.listItems s)


  (* state configuration and maximal orthogonal set *)

  type conf = andstateset 
  type mos  = stateset

  val emptyconf = ANDStateSet.empty
  val emptymos  = StateSet.empty

  fun conf2mos (conf :conf) :mos  = 
	StateSet.fromList $ map InclAND $ ANDStateSet.listItems conf
  fun mos2conf (mos  :mos)  :conf = 
	let fun f (InclAND s) = if null (childrenand s) then s
				else impossible 
				"Mos computed is not a configuration!"
	      | f (InclOR  s) = 
			impossible "Mos computed is not a configuration!"
	in ANDStateSet.fromList $ map f $ StateSet.listItems mos end

  (* computed instances of events and outputs *)

  type einst = event  * value list
  type oinst = output * value list

  (* signal queue *)

  type queue = einst F.fifo

  (* global state between microsteps -- enriched with outputs *)

  type uglobal = conf * venv * denv * queue * oinst list

  (* global state between macrosteps *)

  type mglobal = conf * venv * denv * oinst list

  (* statenv is the stati part of model description. It was implicitly
     global in semantics definition. We make it a tuple which is going to be
     passed to pure functions *)

  type statenv = aor hierarchy * otype env * evgroup env * atr tran list




  (* logical operators *)

  fun vneg v = bool2int $ not $ val2bool v

  fun vand (a :value, b :value) :value = 
	bool2int (val2bool a andalso val2bool b)

  fun vor  (a :value, b :value) :value =
	bool2int (val2bool a orelse val2bool b)



  (* arithmetic negation *)	

  fun aneg (ValI i) = ValI (~i)
    | aneg (ValR r) = ValR (~r)
    | aneg (ValA _) = 
	runtimeErr "Array addresses are illegal as scalars in visualSTATE"




  (* bit operators on integers *)

  fun appb (ope : Word.word*Word.word -> Word.word) (a :int, b :int) :value =
	ValI $ Word.toInt $ ope $ (Word.fromInt a, Word.fromInt b)

  val bxor = appb Word.xorb
  val bor  = appb Word.orb
  val band = appb Word.andb
  val shr  = appb Word.>>
  val shl  = appb Word.<<
	
  fun bneg (ValI i) = ValI $ Word.toInt $ Word.notb $ Word.fromInt i 
    | bneg (ValR r) = 
	runtimeErr "Binary negation expects integer operand. Received real."
    | bneg (ValA _) = 
	runtimeErr "Array addresses are illegal int values in visualSTATE"






  (* evaluate accesses *)	

  fun access (ot : otype env) (acc :access) (venv :venv) : value * oinst list =
	case acc of
		  AccessVar v    => 
			(lookup venv v,[])
		| AccessArr(v,e) => 
			case lookup venv v of
				  ValA a => 
					let val (v1,os) = eval ot e venv
					in (nth (a, getInt v1),os) end
				| _ => runtimeErr $ "Scalar variable " ^ 
					S.name v ^ " accessed as array."



  (* evaluate expressions *)
		
  and eval (ot :otype env) (e :exp) (venv :venv) :value * oinst list = 
	case e of
		  Cst c    => (const c,[])
		
		| Access a => access ot a venv
		
		| Prim1 (ope,e) =>
			let val (v,os1) = eval ot e venv
			in case ope of
				  OpNot   => (vneg v, os1)
				| OpBNot  => (bneg v, os1)
				| OpMinus => (aneg v, os1)
			end  
		| Prim2 (ope,e1,e2) =>
			let val (v1,os1) = eval ot e1 venv
			    val (v2,os2) = eval ot e2 venv
			    val (p1,p2) = promote (v1,v2)
			    val v = case (p1,p2,ope) of
				
				  (ValI a,ValI b,  OpAdd  ) => ValI (a+b)		
				| (ValR a,ValR b,  OpAdd  ) => ValR (a+b)
				
				| (ValI a,ValI b,  OpSub  ) => ValI (a-b)
				| (ValR a,ValR b,  OpSub  ) => ValR (a-b)
				
				| (ValI a,ValI b,  OpMul  ) => ValI (a*b)
				| (ValR a,ValR b,  OpMul  ) => ValR (a*b)
				
				| (ValI a,ValI b,  OpMod  ) => ValI (a mod b)
				| (_,     _,       OpMod  ) => runtimeErr 
					"Modulo used on non integer operands."
				
				| (ValI a,ValI 0,  OpDiv  ) => 
					runtimeErr "Division by zero (0)"
				| (ValI a,ValI b,  OpDiv  ) => ValI (a div b)
				| (ValR a,ValR 0.0,OpDiv  ) =>
					runtimeErr "Division by zero (0.0)"
				| (ValR a,ValR b,  OpDiv  ) => ValR (a/b)

				| (ValI a,ValI b,  OpLt   ) => bool2int(a<b)
				| (ValR a,ValR b,  OpLt   ) => bool2int(a<b)
				
				| (ValI a,ValI b,  OpLte  ) => bool2int(a<=b)
				| (ValR a,ValR b,  OpLte  ) => bool2int(a<=b)

				| (ValI a,ValI b,  OpGt   ) => bool2int(a>b)
				| (ValR a,ValR b,  OpGt   ) => bool2int(a>b)

				| (ValI a,ValI b,  OpGte  ) => bool2int(a>=b)
				| (ValR a,ValR b,  OpGte  ) => bool2int(a>=b)
	
				| (ValI a,ValI b,  OpEq   ) => bool2int(a=b)
				| (ValR a,ValR b,  OpEq   ) => bool2int(a=b)

				| (ValI a,ValI b,  OpNeq  ) => bool2int(a<>b)
				| (ValR a,ValR b,  OpNeq  ) => bool2int(a<>b)

				| (ValI a,ValI b,  OpBXor ) => bxor (a,b)
				| (ValR a,ValR b,  OpBXor ) => runtimeErr
					"Binary xor used on nonint operands"

				| (     a,     b,  OpAnd  ) => vand (a,b)
				| (     a,     b,  OpOr   ) => vor  (a,b)

				| (ValI a,ValI b,  OpBAnd ) => band (a,b)
				| (ValR a,ValR b,  OpBAnd ) => runtimeErr
					"Binary AND used on nonint operands."
				
				| (ValI a,ValI b,  OpBOr  ) => bor  (a,b)
				| (ValR a,ValR b,  OpBOr  ) => runtimeErr
					"Binary OR used on nonint operands."

				| (ValI a,ValI b,  OpShr  ) => shr (a,b)
				| (ValR a,ValR b,  OpShr  ) => runtimeErr
					"Binary SHR used on nonint operands."

				| (ValI a,ValI b,  OpShl  ) => shl (a,b)
				| (ValR a,ValR b,  OpShl  ) => runtimeErr
					"Binary SHL used on nonint operands."

				| _ => impossible "Unexpected type mismatch."	
				(* but note that typechecker is nonexistent *)
	
			in (v, os1 @ os2) end

		| Call oe => 
			let val os = oeval ot oe venv
			    val v  = defvalty $ #1 $ lookup ot (#1 oe)
			in (v,os) end
		| After (e1,e2) => 
			let val (_,os1) = eval ot e1 venv
			    val (v,os2) = eval ot e2 venv
			in (v, os1 @ os2) end
			



  (* evaluate output expressions *)
  
  and oeval (ot :otype env) (oe :oexp) (venv :venv) : oinst list =
	let val vsos    = map (fn pe => eval ot pe venv) (#2 oe)
	    val (vs,os) = ListPair.unzip vsos
	in concat os @ [(#1 oe,vs)] end

  (* evaluate signal expressions *)

  and seval (ot :otype env) (se :sexp) (venv :venv) : einst * oinst list =
	let val ssos    = map (fn pe => eval ot pe venv) (#2 se)
	    val (ss,os) = ListPair.unzip ssos
	in ((#1 se,ss), concat os) end




  (* now "specialize" to obtain boolean and arithmetic evaluators *)

  val aeval = eval

  fun beval (ot :otype env) (e :exp) (venv :venv) : bool * oinst list = 
	let val (res,os) = eval ot e venv 
	in (val2bool res,os) end



  (* evaluation of assignments *)

  fun vector_update (ValA vs :value) (ValI pos :value) (new :value) :value =
	let val pref = if pos > 0 then take (vs, pos-1) else []
	    val suff = drop (vs,pos)
	in ValA (pref @ new :: suff) end
    | vector_update v (ValI _) _ = runtimeErr "Indexing non array."
    | vector_update _ _        _ = runtimeErr "Noninteger array index."



  fun asgn (ot :otype env) (a :assgn) (venv :venv) (os :oinst list) 
							:venv * oinst list =
	case a of
		  (AccessVar v,e) => 
			let val (ev,os1) = eval ot e venv
			in (bind1 venv (v,ev), revAppend(os1,os)) end
		| (AccessArr (a,e1),e2) =>
			let 
				val (v1,os1) = aeval ot e1 venv 
				val (v2,os2) = aeval ot e2 venv 
				val arr      = lookup venv a
				val narr     = vector_update arr v1 v2
				val os3      = revAppend(os2,revAppend(os1,os))
			in (bind1 venv (a,narr),os3) end
		
  (* Action execution *)	

  fun exec_cmd (ot :otype env) (cmd :command,(venv,os)) : venv * oinst list =
	case cmd of
		  Expr(Call oe) => (venv, revAppend(oeval ot oe venv, os))
		| Expr e        => (venv,revAppend(#2(eval ot e venv),os))
		| Assgn a       => asgn ot a venv os

  fun exec (ot :otype env) ((cmds,sigs) :action) (venv :venv) 
		(q :queue) (os :oinst list)
			: venv * queue * oinst list =
	let	
		val (venv1,os1) = foldl (exec_cmd ot) (venv,os) cmds
		fun sev (se,(q,os)) =
			let val (si,os1) = seval ot se venv1
			in (F.enqueue (q,si), revAppend (os1,os)) end
		val (q1,os2) = foldl sev (q,os1) sigs
	in (venv1, q1, os2) end handle e => intErr e 200302181


  (* exiting a scope *)	

  fun exitand (ot :otype env) (s :aor andstate) (mos :mos) (venv :venv) 
	(denv :denv) (q :queue)  (os :oinst list) 
		: venv * denv * queue * oinst list = 

	if StateSet.equal( mos, StateSet.singleton (InclAND s)) 
	then (venv,denv,q,os)
	else let
		val _ = assert (fn () => not $ StateSet.member(mos,InclAND s)) 
				"See figure 2/p.10."

		fun fe (si :aor orstate,(venv,denv,q,os)) =  
			let val desi= descendor si
			    val mosi= StateSet.intersection (mos,descendor si)
			    val (venvi,denvi,qi,osi) = 
				exitor ot si mosi venv denv q os
			in (venvi,denvi,qi, osi) end
		val (venv1,denv1,q1,os1) = 
			foldl fe (venv,denv,q,os) (childrenand s)

	in (venv1, denv1, q1, os1) end

  and exitor (ot :otype env) (s :aor orstate) (mos :mos) (venv :venv) 
	(denv :denv) (q :queue) (os :oinst list)
		: venv * denv * queue * oinst list = 
	let
		fun ac (st :aor andstate) = StateSet.isSubset (mos,descendandst st) 
		val sopt = find ac (childrenor s)

		val _ = assert (fn () => isSome sopt) 
				"Attempt to exit non active state."
				
		val s' = valOf sopt
		val (venv1,denv1,q1,os1) = exitand ot s' mos venv denv q os
		val (venv2,q2,os2) = exec ot (exand s') venv1 q1 os1
		val denv2 = if ishistoryor s then 
				bind1 denv1 (nameor s,nameand s') 
			    else denv1
	in (venv2,denv2,q2,os2) end


  (* entering a scope *)

  fun enterand (h :aor hierarchy) (ot :otype env) (tgt :stateset) 
	(s :aor andstate) (venv :venv) (denv :denv) (q :queue) 
		(os :oinst list) :conf * venv * queue * oinst list =

	if StateSet.isEmpty tgt andalso null (childrenand s) then 
		(ANDStateSet.singleton s, venv, q, os) 
	else let
		val _ = assert (fn ()=> StateSet.isSubset (tgt,descendandst s))
				"No target in scope being entered: rule 2."

		fun fe (si :aor orstate,(conf,venv,q,os)) = 
			let val desi = descendor si
			    val tgti = StateSet.intersection (tgt,descendor si)
			    val (confi,venvi,qi,osi) = 
				enteror h ot tgti si venv denv q os
			in (ANDStateSet.union (conf,confi),venvi,qi,osi) end
		
		val kids = childrenand s
		val (conf1,venv1,q1,os1) = 
			foldr fe (emptyconf,venv,q,os) kids

	in (conf1, venv1, q1, os1) end

  and enteror (h :aor hierarchy) (ot :otype env) (tgt :stateset) (s :aor orstate) 
	(venv :venv) (denv :denv) (q :queue) 
		(os :oinst list) :conf * venv * queue * oinst list =
	if StateSet.isEmpty tgt then
		let val s' = getandh h (lookup denv (nameor s))
		    val (venv1,q1,os1) = exec ot (enand s') venv q os
		in enterand h ot emptymos s' venv1 denv q1 os1 end
	else 
		let 
		    val _ = assert (fn()=>StateSet.isSubset(tgt,descendorst s))
				"No target in scope being entered: rule 3."
		    val kids = childrenor s 
		    val lca  = LCA h $ map name $ StateSet.listItems tgt
		    val anc  = ancesthst h lca
		    val sset = Set.intersection (* FIXME: does it pay off? *)
				(Set.fromList $ map nameand kids, 
				 Set.fromList anc)
		    val _    = assert (fn () => Set.numItems sset = 1)	
				"Non unique targets ancestor:rule 3."
		    
		    val ssym = hd (Set.listItems sset)
		    val s'   = valOf $ find (fn s => nameand s = ssym) kids
			       handle Option.Option => impossible "should be a child"
		    val (venv1,q1,os1) = exec ot (enand s') venv q os
		    val tgt' = StateSet.difference1 (tgt, InclAND s')
		in enterand h ot tgt' s' venv1 denv q1 os1 end


  (* firing a transition *)		

  (* check if is is not covered (ancestrally realted) 
     to any state in scopes *)

  fun covered (h :aor hierarchy) (scopes :orstateset) (is :aor orstate) :bool =
	let val nis = nameor is
	in ORStateSet.exists (fn s=>ancestorh h (nameor s) nis) scopes end


  (* implicit scope of state change. mos is current configuration
     passed as mos for efficiency reasons *)

  fun iscope (h :aor hierarchy) (mos :mos) (tgt :aor andstate) :aor orstate =
	let
		val p       = valOf $ getor $ getstate h (parentand tgt)
		val desc    = descendor p
		val overlap = StateSet.intersection (mos, desc)
	in
		if StateSet.isEmpty overlap 
		then iscope h mos (valOf $ getand $ getstate h (parentor p)) 
		else p
	end handle Option.Option => impossible "implicit scope is not an orstate!"

  (* generalized scope of transition *)

  fun iscopes (h :aor hierarchy) (conf :conf) (t :atr tran) :aor orstate list =	
	map (iscope h (conf2mos conf)) (map (getandh h) (#targets t))
	(* a silly rewrite with this gethand. Might be faster to change
	   the return type to orstate list, but this should b carefully
	   examined *)



  (* implicit scope of t is passed as parameter for efficiency - its known at
     that point in caller *)

  fun fire (h :aor hierarchy) (ot :otype env) (t :atr tran) (iss :aor orstate list) 
	((conf,venv,denv,q,os) :uglobal) :uglobal =
	let 
		val mos  = conf2mos conf

		(* abandon type of targets for efficiency *)
		val tgts = StateSet.fromList $ map (getstate h) (#targets t)
		
		fun ex (s :aor orstate,(venv,denv,q,os)) = 
			let val desc = descendor s
			    val mos1 = StateSet.intersection (desc,mos)
			in exitor ot s mos1 venv denv q os end
		
		val (venv1,denv1,q1,os1) =
			foldl ex (venv,denv,q,os) iss

		val (venv2,q2,os2) = exec ot (#action t) venv1 q1 os1

		fun en (si :aor orstate,(conf,venv,q,os)) = 
			let val desci = descendor si
			    val tgtsi = StateSet.intersection (desci,tgts)
			    val (conf1,venv1,q1,os1) = 
				enteror h ot tgtsi si venv denv1 q os 
			    val conf2 = ANDStateSet.difference (conf,onlyand desci)
			    val conf3 = ANDStateSet.union (conf2,conf1)
			in (conf3,venv1,q1,os1) end

		val (conf1,venv3,q3,os3) = foldl en (conf,venv2,q2,os2) iss

	in (conf1,venv2,denv1,q2,os3) end
	
	(* FIXME-PROBLEM nonresolvable transition conflicts will give 
	   rise to some very criptic runtime or internal errors *)




  (* active transitions *)	

  (* mos is current configuration. This is not passed as conf
     for simple efficiency gain. FIXME: algoritm of notinstate is 
     ineffcient *)

    
  fun instate (h :'a hierarchy) (mos :mos) (s :stateid) :bool = 
	StateSet.exists (fn m=> ancestorhst h s (name m)) mos
 
  fun notinstate (h :'a hierarchy) (mos :mos) (s :stateid) :bool = 
	not (instate h mos s)



  fun active (h:'a hierarchy) (ot :otype env) (eg :evgroup env) (ei :einst) (t :atr tran) 
	(conf :conf) (venv :venv) (os :oinst list) :  bool * oinst list =
	let
		val e = #1 ei
		val mos = conf2mos conf

		val venv1 = case #trigger t of
				  EventGroup g => venv
				| Event (e,params) =>
					bindZip venv (params,#2 ei)

		fun allguards (g::gs) os =
			let val (res,os1) = beval ot g venv1
			    val os2       = revAppend(os1,os)
			in if res then allguards gs os2 else (false,os2) end
		  | allguards [] os = (true,os)
	in
		if Set.member (events eg t, e) 
			andalso all (instate h mos) (#pos t)
			andalso all (notinstate h mos) (#neg t)
		then allguards (#guard t) os
		else (false,os)
	end
	



  (* Microstep (includes conflict resolution) *) 

  fun micro ((h,ot,eg,trans) :statenv) (ei :einst) 
	((conf,venv,denv,q,os) :uglobal) : uglobal = 
	let
		fun ac (t,(alst,os)) =
			let val (a,os1) = active h ot eg ei t conf venv os 
			in if a then (t::alst,os1) else (alst,os1) end

		val (atrans,os1) = foldl ac ([],os) trans (* active trans   *)

	
		fun bindparams (venv :venv) (t :atr tran) :venv * value list =
			case #trigger t of
				  EventGroup _ => (venv,[])
				| Event (e,params) =>
					(bindZip venv (params,#2 ei),
					 map (lookup venv) params)

		fun popparams (venv :venv) (t :atr tran) 
			(saved :value list) :venv =
			case #trigger t of
				  EventGroup _ => venv
				| Event (e,params) =>
					bindZip venv (params,saved)
		
		(* This is a particularly nasty hack. The problem lies in
		   supposedly uniform semantics of statecharts. See example 
		   trans03.  very unefficient: this approach should be avoided 
		   in interpreter for generated code *)

		fun uncover (s :aor orstate) (iss :aor orstate list) :aor orstate list =
			if exists (fn is => ancestorhst h (nameor is) (nameor s)) iss 
			then iss
			else s :: List.filter (fn is => not $ 
					ancestorhst h (nameor s) (nameor is)) iss
	
		(* ft - attempt to fire single transition *)
		fun ft (t:atr tran,((conf,venv,denv,q,os),scopes)) = 
			let
				val iss = uncover (getorh h (#scope t)) 
							(iscopes h conf t)
				val _   = assert (fn ()=> not $ exists 
					  (fn is=> ORStateSet.member (scopes,is)) iss)
					  "Conflict has not been fully resolved."
			in
				(* skip transition in resolvable conflict *)
				
				if exists (covered h scopes) iss then
					((conf,venv,denv,q,os),scopes)
					
				(* fire transition *)

				else let
					(* pushing params is painful with
					   this modelling of vars *)
					val (venv1,saved) = bindparams venv t
					val (conf1,venv1,denv1,q1,os1) =
						fire h ot t iss 
							(conf,venv1,denv,q,os)
					val venv2 = popparams venv1 t saved
				in 
					((conf1,venv2,denv1,q1,os1), 
					 ORStateSet.addList(scopes,iss) )
				end
			end		

		val (uglobal,scopes) = foldl ft 
			((conf,venv,denv,q,os1),ORStateSet.empty) atrans

	in uglobal end


  (* macro step *)

  fun macro (statenv :statenv) ((conf,venv,denv,q,os) :uglobal) :mglobal = 
	let val (q1,ei) = F.dequeue q
	in macro statenv $ micro statenv ei (conf,venv,denv,q1,os)
	end handle Dequeue => (conf,venv,denv,os)
			
	

  (* interface for feeding events *)

  fun step (sys : (aor,atr) system) (ei :einst) ((conf,venv,denv,os) :mglobal) :mglobal = 
	let val statenv = 
		(#hierarchy sys, #otypes sys, #evgroups sys, #trans sys)
	in macro statenv (conf,venv,denv,F.enqueue (F.empty,ei),os) end



  (* initialization step for system *)
  
  fun init ({hierarchy,evgroups,otypes,denv,venv,trans,...} :(aor,atr) system) :mglobal =
	let 
		val statenv = (hierarchy, otypes, evgroups, trans)
		val (conf1,venv1,q1,os1) = 
			enterand hierarchy otypes emptymos (gettree hierarchy) 
			venv denv F.empty []
	in 
		macro statenv (conf1,venv1,denv,q1,os1)
	end handle e => intErr e 200302180
		
end  
