(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/parse.sml,v
   Comment: Wrapper for vsr/vsp parser. 
            File originates from Peter Sestoft's example in MoscowML distriubtion.
            Adopted by Jorn Lind Nielsem for his converter of VS models. 
            Finally adopted by AW for the scope code generator. There is hardly
            any resemblance of the previous versions and current one.
   Release: scope-0-15_final

     Copyright (C) Peter Sestoft
     Copyright (C) until 2001 Jorn Lind Nielsen
     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   parse.sml,v 1.14 2003/12/15 22:07:55 wasowski Exp
*)
structure parse :> parse  = struct

  open Lexing Parsing Err Say

  structure V = vsr
  structure L = Location

  (* Create lexer from file (given by path) *)

  fun createLexerFile fpath =
	let val is = Nonstdio.open_in_bin fpath
	    val lexbuf = createLexer 
		(fn buf => fn n => Nonstdio.buff_input is buf 0 n) 
	in (is,lexbuf) end
	

  (* Parse; show offending program piece on error *)

  fun parseReport path lexbuf = 
	(parser.File lexer.Token lexbuf before clearParser ()
	 handle ParseError _ =>
		   let val a = getLexemeStart lexbuf
		       val b = getLexemeEnd lexbuf
		   in syntaxErr (L.Loc(a,b)) (SOME path) "Syntax error." end)
	handle ScopeExn i => finalize clearParser (ScopeExn(addFile path i)) 
	     |        exn => finalize clearParser exn

	     
  (* SOME ("nice" stuff) to handle bad grammar of vsp in postprocessing *) 

  fun processFieldDir lst =
	let fun f ((V.SysName  a)::t, fs,cs,ns) = f (t, fs,    cs,    a::ns)
	      | f ((V.FilePath a)::t, fs,cs,ns) = f (t, a::fs, cs,    ns   )
	      | f ((V.CallTag  a)::t, fs,cs,ns) = f (t, fs,    a::cs, ns   )
	      | f (_::t,              fs,cs,ns) = f (t, fs,    cs,    ns   )
	      | f ([],                fs,cs,ns) =   (fs, cs, ns )
	in f (lst,[],[],[]) end

  (* first form of pattern should always match with this parser *)

  fun extractModuleInfo [_,_,_,V.Dir_Field( SOME lst )] = processFieldDir lst
    | extractModuleInfo _ = Err.impossible ""

  (* FIXME: more specific exception should be caught *)

  fun isFile (name : string) : bool =
	not (FileSys.isDir name) handle _ => false


  (* returns a name if it is accessible in current directory.
     or name concatenated with path if the file can be found there.
     Raises fail otherwise. *)

  fun findfile (name:string) (NONE:string option) = 
	if isFile name then name 
	else raise Fail ("Error openning file: " ^ name)
    | findfile name (SOME path) =
	if isFile (path ^ "/" ^ name) 
	then path ^ "/" ^ name else findfile name NONE
	
  (* parse dependent systems (vsp and vsr files) *)

  (* Parse a program or a project from a file *)			
                                                                           
  fun parse f project_path =                                               
      let
          val path        = findfile f project_path                             
          val (is,lexbuf) = createLexerFile path                                
          fun destr ()    = BasicIO.close_in is 
	  val expr        = parseReport path lexbuf 
			    handle e => finalize destr e

      in expand f expr before destr () end


  (* pulls out some info from tags to ast global level *)
  (* FIXME => Does this simple function need to be so big? *)
	
	and expand f ( V.Program { directives, interface,
				   machines, top, filepath } ) = 
			V.Program {  

				directives = directives, 
				interface  = interface,
				machines   = machines, 
				top        = top, 
				filepath   = Symbol.symbol f
			}

	  | expand f (V.Project { directives, interface, files, 
				  programs, calls, sysnames, filepath } ) = 
		let
			val   (fs,cs,ns) = extractModuleInfo directives
			val project_path = SOME(Path.dir f)
		in      
			V.Project {  

				directives = directives, 
				interface  = interface,
				files      = fs, 
				calls      = cs, 
				sysnames   = ns,
				programs   = map (compFile project_path) fs, 
				filepath   = Symbol.symbol f
			}
		end

	and compFile project_path (n,m,s) = 
		let val name = Symbol.name s
		in case parse name project_path of
			  V.Project p   => raise Fail "Projects may not incorporate another projects!"
			| V.Program prg => (n,m,prg)
		end

			 
end
